package ch.uzh.csg.filesharing;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigInteger;
import java.net.InetSocketAddress;
import java.net.URISyntaxException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Random;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.logging.LogManager;

import org.apache.mina.core.session.IoSession;
import org.apache.mina.util.ConcurrentHashSet;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.uzh.csg.filesharing.fileutils.FileChunkMapping;
import ch.uzh.csg.filesharing.fileutils.FileChunks;
import ch.uzh.csg.filesharing.fileutils.FileUtils;
import ch.uzh.csg.filesharing.fileutils.RandomBitSet;
import ch.uzh.csg.filesharing.fileutils.SplitterUtils;
import ch.uzh.csg.filesharing.future.FutureCheck;
import ch.uzh.csg.filesharing.future.FutureCheckImpl;
import ch.uzh.csg.filesharing.future.FutureFile;
import ch.uzh.csg.filesharing.future.FutureFileChunk;
import ch.uzh.csg.filesharing.future.FutureFileChunkImpl;
import ch.uzh.csg.filesharing.future.FutureFileImpl;
import ch.uzh.csg.filesharing.future.FutureGenericFile;
import ch.uzh.csg.filesharing.future.FutureGenericFileImpl;
import ch.uzh.csg.filesharing.future.FuturePublish;
import ch.uzh.csg.filesharing.future.FuturePublishImpl;
import ch.uzh.csg.filesharing.future.FutureURN;
import ch.uzh.csg.filesharing.future.FutureURNImpl;
import ch.uzh.csg.filesharing.future.Progress;
import ch.uzh.csg.tomp2p.Node;
import ch.uzh.csg.tomp2p.Replication;
import ch.uzh.csg.tomp2p.Utils;
import ch.uzh.csg.tomp2p.action.FutureDHT;
import ch.uzh.csg.tomp2p.action.FutureTracker;
import ch.uzh.csg.tomp2p.asyncloop.Loopable;
import ch.uzh.csg.tomp2p.asyncloop.Recursive;
import ch.uzh.csg.tomp2p.common.BaseFutureAdapter;
import ch.uzh.csg.tomp2p.common.FutureResponse;
import ch.uzh.csg.tomp2p.connection.BindInformation;
import ch.uzh.csg.tomp2p.message.Message;
import ch.uzh.csg.tomp2p.routing.NodeAddress;
import ch.uzh.csg.tomp2p.rpc.ReplyFilter;
import ch.uzh.csg.tomp2p.rpc.StoreMultiMap;
import ch.uzh.csg.tomp2p.storage.ByteArrayObject;

public class FileSharingNode extends Node
{
	final private static int PARALLEL = 1;
	public static enum Strategy
	{
		TFT, PSH2, PSHR, NONE
	};
	final private Strategy currentStrategy;
	final private File publishDir, incomingDir;
	public final static String SEEDER = "SEEDER";
	public final static String FILE_NAME = "FILE_NAME";
	public final static String TRACKERS = "TRACKERS";
	private FileChunkMapping fileChunkMapping;
	final private static Logger logger = LoggerFactory.getLogger(FileSharingNode.class);
	final private History history;
	private FileSharingRPC fsRCP;
	private boolean freeriding = false;
	final private Set<URN> publishedURN = new ConcurrentHashSet<URN>();
	private final int splitSize;
	private final static Random rnd = new Random();
	private final ListenData listenData = new ListenData();
	private final List<String> files;
	private final List<URN> urns;
	final private Map<NodeAddress, Long> noNeededParts = new ConcurrentHashMap<NodeAddress, Long>();
	final private Map<NodeAddress, Map<URN, TimeBitSet>> chunkAvailability = new ConcurrentHashMap<NodeAddress, Map<URN, TimeBitSet>>();
	// final private ExpiringMap<NodeAddress, Long> noSlot = new
	// ExpiringMap<NodeAddress, Long>(30, 10);
	final private Map<NodeAddress, Long> noIntermediatePeers = new ConcurrentHashMap<NodeAddress, Long>();
	final private double goodwill;
	static
	{
		try
		{
			InputStream is = FileSharingNode.class
					.getResourceAsStream("/filesharinglog.properties");
			if (is != null)
			{
				LogManager.getLogManager().readConfiguration(is);
			}
		}
		catch (SecurityException e)
		{
			e.printStackTrace();
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
	}

	public static void dispose()
	{
		Node.dispose();
		SplitterUtils.getThreadPool().shutdown();
	}

	public FileSharingNode(int p2pId, BigInteger nodeId, Strategy currentStrategy, int credit,
			int splitSize, List<String> files, List<URN> urns, File publishDir, File incomingDir,
			int timeToTrade, double goodwill) throws IOException
	{
		super(p2pId, nodeId);
		this.currentStrategy = currentStrategy;
		this.splitSize = splitSize;
		history = new History(splitSize, credit, timeToTrade);
		this.files = files;
		this.urns = urns;
		this.publishDir = publishDir;
		this.incomingDir = incomingDir;
		this.goodwill = goodwill;
		createLocalDir();
	}

	private void createLocalDir() throws IOException
	{
		if (!publishDir.exists())
			publishDir.mkdirs();
		if (!incomingDir.exists())
			incomingDir.mkdirs();
	}

	public File getPublishDir()
	{
		return publishDir;
	}

	@Override
	public void listen(int udpPort, int tcpPort, BindInformation bindInformation,
			Replication replication, Object statServer) throws IOException
	{
		super.listen(udpPort, tcpPort, bindInformation, replication, statServer);
		afterListen();
	}

	public FileSharingRPC getFileSharingRPC()
	{
		if (fsRCP == null)
			throw new RuntimeException("Not listening to anything. Listen first");
		return fsRCP;
	}

	public boolean isPartially(URN urn)
	{
		return fileChunkMapping.isPartially(urn);
	}

	public boolean isComplete(URN urn)
	{
		return fileChunkMapping.isComplete(urn);
	}

	public File getCompleteFile(URN urn)
	{
		return fileChunkMapping.getCompleteFile(urn);
	}

	public void resetChunkCounter()
	{
		for (BigInteger key : fileChunkMapping.getFileMapping().keySet())
		{
			FileChunks value = fileChunkMapping.getFileMapping().get(key);
			if (value != null)
			{
				value.resetChunkCounter();
			}
		}
	}

	public int countChunks()
	{
		int chunks = 0;
		for (BigInteger key : fileChunkMapping.getFileMapping().keySet())
		{
			FileChunks value = fileChunkMapping.getFileMapping().get(key);
			if (value != null)
			{
				chunks += value.getChunkCounter();
			}
		}
		return chunks;
	}

	public FuturePublish addFileAndPublish(File file) throws IOException, URISyntaxException
	{
		return addFileAndPublish(file, new URN(file));
	}

	public FuturePublish addFileAndPublish(File file, URN urn) throws IOException,
			URISyntaxException
	{
		logger.debug("publish file (name)" + file.getName() + ", I'm " + getNodeID());
		file.getParentFile().mkdirs();
		fileChunkMapping.addFileMapping(file, urn);
		return publishURN(urn);
	}

	public FuturePublish publishURN(final URN urn)
	{
		final FuturePublish futurePublish = new FuturePublishImpl();
		if (freeriding || !isListening())
		{
			String reason = freeriding ? "This node is a freerider"
					: "This node is not listening to a port";
			futurePublish.failed(Message.TYPE_NOT_OK, reason);
			return futurePublish;
		}
		logger.debug("publish file or chunk (" + urn + "), key=" + urn.getHash() + ". I'm node "
				+ getNodeID());
		// TODO: search by name, FastSS
		add(Utils.makeSHAHash(urn.getName()), FILE_NAME, urn).addListener(
				new BaseFutureAdapter<FutureDHT>()
				{
					@Override
					public void operationComplete(FutureDHT future)
					{
						futurePublish.setNamePublished(future.isSuccess());
					}
				});
		// search for a tracker
		get(urn.getHash(), TRACKERS).addListener(new BaseFutureAdapter<FutureDHT>()
		{
			@Override
			public void operationComplete(FutureDHT future)
			{
				// if we have more than 0 trackers, attach to a tracker,
				// otherwise
				// create new one
				if (future.isSuccess() && future.getData().size() > 0)
				{
					getTracker().add((NodeAddress) future.getData().iterator().next(),
							urn.getHash(), SEEDER, getNodeAddress()).addListener(
							new BaseFutureAdapter<FutureTracker<NodeAddress>>()
							{
								@Override
								public void operationComplete(FutureTracker<NodeAddress> future)
								{
									futurePublish.setTrackerPublished(future.isSuccess());
								}
							});
					// TODO: check if tracker is full
				}
				else
				{
					// there is no tracker or not enough trackers, create and
					// add one
					// tracker
					add(urn.getHash(), TRACKERS, getNodeAddress()).addListener(
							new BaseFutureAdapter<FutureDHT>()
							{
								@Override
								public void operationComplete(FutureDHT future)
								{
									futurePublish.setTrackerPublished(future.isSuccess());
								}
							});
					getTracker().addLocal(urn.getHash(), SEEDER, getNodeAddress());
				}
			}
		});
		return futurePublish;
	}

	public FutureURN searchURN(final String name)
	{
		final FutureURN futureURN = new FutureURNImpl();
		final BigInteger nameHash = Utils.makeSHAHash(name);
		logger.debug("Search for file (name,nameHash,domain): (" + name + "," + nameHash + ","
				+ FILE_NAME + "), resolve first");
		get(nameHash, FILE_NAME).addListener(new BaseFutureAdapter<FutureDHT>()
		{
			@Override
			public void operationComplete(FutureDHT futureDHT)
			{
				try
				{
					if (!futureDHT.isFailed() && futureDHT.getData().size() > 0)
					{
						logger.debug("Serch for URN with name: " + name + " found");
						Collection<URN> tmp = new ArrayList<URN>();
						for (Object obj : futureDHT.getData())
						{
							if (obj instanceof URN)
								tmp.add((URN) obj);
						}
						futureURN.setURNs(tmp);
					}
					else
					{
						logger.warn("file: " + name + " not found for file here, stop: ("
								+ nameHash + ")" + futureDHT.isFailed());
						futureURN.failed(Message.TYPE_NOT_OK, "file: " + name
								+ " not found for file here, stop: ("
								+ nameHash + ")" + futureDHT.isFailed());
					}
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		});
		return futureURN;
	}

	public FutureFile downloadURN(final URN urn)
	{
		return downloadURN(urn, Integer.MAX_VALUE);
	}
	private Thread downloader;

	public FutureFile downloadURN(final URN urn, final int timeout)
	{
		final FutureFile futureFile = new FutureFileImpl(urn, fileChunkMapping);
		final Collection<NodeAddress> peerThatHaveChunks = new ConcurrentHashSet<NodeAddress>();
		downloader = new Thread(downloadFromPeers2(urn, peerThatHaveChunks, futureFile, timeout));
		downloader.start();
		return futureFile;
	}

	@Override
	public synchronized void shutdown()
	{
		super.shutdown();
		if (downloader != null)
			downloader.interrupt();
	}

	private void checkForPeersWithChunks(final URN urn,
			final Collection<NodeAddress> peerThatHaveChunks)
	{
		logger.debug("checker run2");
		FutureDHT futureDHT = get(urn.getHash(), TRACKERS);
		futureDHT.awaitUninterruptibly();
		if (futureDHT.isSuccess())
		{
			for (Object bao : futureDHT.getData())
			{
				FutureTracker<NodeAddress> futureTracker = getTracker().get((NodeAddress) bao,
						urn.getHash(), SEEDER);
				futureTracker.awaitUninterruptibly();
				if (futureTracker.isSuccess())
				{
					for (NodeAddress peerWithChunk : futureTracker.getChildTrackers())
					{
						boolean changed = false;
						if (!peerThatHaveChunks.contains(peerWithChunk))
						{
							peerThatHaveChunks.add(peerWithChunk);
							changed = true;
						}
						if (changed)
						{
							history.notifyChange();
						}
					}
				}
				else
				{
					System.err.println("seeders not found");
				}
			}
		}
		else
		{
			logger.debug("FutureDHT in checekr faild " + futureDHT.getFailedReason());
		}
		for (Iterator<NodeAddress> i = noNeededParts.keySet().iterator(); i.hasNext();)
		{
			NodeAddress key = i.next();
			Long tmp = noNeededParts.get(key);
			if (tmp != null && System.currentTimeMillis() - tmp > 30 * 1000)
				i.remove();
		}
		for (Iterator<NodeAddress> i = noIntermediatePeers.keySet().iterator(); i.hasNext();)
		{
			NodeAddress key = i.next();
			Long tmp = noIntermediatePeers.get(key);
			if (tmp != null && System.currentTimeMillis() - tmp > 10 * 1000)
				i.remove();
		}
		for (Iterator<NodeAddress> i = chunkAvailability.keySet().iterator(); i.hasNext();)
		{
			NodeAddress key = i.next();
			Map<URN, TimeBitSet> tmp1 = chunkAvailability.get(key);
			if (tmp1 != null)
			{
				for (Iterator<URN> i1 = tmp1.keySet().iterator(); i1.hasNext();)
				{
					URN u = i1.next();
					TimeBitSet s = tmp1.get(u);
					if (s != null && System.currentTimeMillis() - s.getTime() > 30 * 1000)
						i1.remove();
				}
			}
		}
	}

	private Runnable downloadFromPeers2(final URN urn,
			final Collection<NodeAddress> peerThatHaveChunks, final FutureFile futureFile,
			final int timeout)
	{
		final long start = System.currentTimeMillis();
		return new Runnable()
		{
			@Override
			public void run()
			{
				try
				{
					long check = 0;
					System.err.println("download! " + urn);
					while (!futureFile.isCompleted() && !downloader.isInterrupted()
							&& (System.currentTimeMillis() - start) < timeout)
					{
						if (System.currentTimeMillis() - check > 10 * 1000)
						{
							checkForPeersWithChunks(urn, peerThatHaveChunks);
							System.err.println("check found with " + peerThatHaveChunks.size());
							check = System.currentTimeMillis();
						}
						FutureGenericFile fgf = incentiveScheme(urn, peerThatHaveChunks, futureFile);
						fgf.awaitUninterruptibly();
						if (fgf.isSuccess())
						{
							logger.debug("Downloaded URN successfull, we have the file! " + urn
									+ " / " + getNodeID());
							// futureFile.addDebug("Downloaded URN successfull, we have the file! "
							// + urn + " / " + getNodeID());
							futureFile.setFile(fgf.getFile());
						}
						else
						{
							logger.debug("cannot start download " + fgf.getFailedReason());
							System.err.println("cannot start download " + fgf.getFailedReason());
							Utils.sleep(5000);
						}
					}
				}
				finally
				{
					futureFile.failed(Message.TYPE_NOT_OK, "Timeout");
					// checker = null;
					// downloader = null;
				}
			}
		};
	}

	private FutureGenericFile incentiveScheme(final URN urn,
			final Collection<NodeAddress> peerThatHaveChunks, final FutureFile futureFile)
	{
		if (currentStrategy == Strategy.NONE)
		{
			for (NodeAddress downloadFrom : peerThatHaveChunks)
			{
				if (!noNeededParts.containsKey(downloadFrom))
				{
					return downloadFileChunksFromPeer(urn, downloadFrom, futureFile, false);
				}
			}
		}
		if (currentStrategy == Strategy.TFT || currentStrategy == Strategy.PSH2
				|| currentStrategy == Strategy.PSHR)
		{
			List<NodeAddress> tmp = new ArrayList<NodeAddress>();
			for (NodeAddress downloadFrom : peerThatHaveChunks)
			{
				if (!noNeededParts.containsKey(downloadFrom)
						&& history.canDownloadFrom(downloadFrom, splitSize))
					tmp.add(downloadFrom);
			}
			while (tmp.size() > 0)
			{
				NodeAddress na = tmp.remove(rnd.nextInt(tmp.size()));
				return downloadFileChunksFromPeer(urn, na, futureFile, false);
			}
		}
		if (currentStrategy == Strategy.TFT)
		{
			List<NodeAddress> tmp = new ArrayList<NodeAddress>();
			for (NodeAddress downloadFrom : peerThatHaveChunks)
			{
				if (!noNeededParts.containsKey(downloadFrom))
					tmp.add(downloadFrom);
			}
			if (tmp.size() > 0)
			{
				NodeAddress downloadFrom = tmp.get(rnd.nextInt(tmp.size()));
				return downloadFileChunksFromPeer(urn, downloadFrom, futureFile, false);
			}
		}
		if (currentStrategy == Strategy.PSH2 || currentStrategy == Strategy.PSHR)
		{
			// PSH
			List<NodeAddress> tmp = new ArrayList<NodeAddress>();
			for (NodeAddress downloadFrom : peerThatHaveChunks)
			{
				if (!noNeededParts.containsKey(downloadFrom))
					tmp.add(downloadFrom);
			}
			if (tmp.size() > 0)
			{
				NodeAddress downloadFrom = tmp.get(rnd.nextInt(tmp.size()));
				return downloadFileChunksFromPeer(urn, downloadFrom, futureFile, true);
			}
		}
		// System.err.println("There are no peers we can download from");
		FutureGenericFile ffg = new FutureGenericFileImpl();
		ffg.failed(Message.TYPE_NOT_OK, "no match");
		return ffg;
	}

	private BitSet getChunkAv(NodeAddress downloadFrom, URN urn)
	{
		Map<URN, TimeBitSet> tmp = chunkAvailability.get(downloadFrom);
		if (tmp != null)
			return tmp.get(urn);
		return null;
	}

	private void putChunkAv(NodeAddress downloadFrom, URN urn, BitSet chunks)
	{
		Map<URN, TimeBitSet> tmp = chunkAvailability.get(downloadFrom);
		if (tmp == null)
		{
			tmp = new ConcurrentHashMap<URN, TimeBitSet>();
			chunkAvailability.put(downloadFrom, tmp);
		}
		tmp.put(urn, new TimeBitSet(chunks));
	}

	private FutureGenericFile downloadFileChunksFromPeer(final URN urn,
			final NodeAddress downloadFrom, final Progress progress, final boolean psh)
	{
		logger.debug("Download chunks from " + downloadFrom + " for URN " + urn + " / "
				+ getNodeID());
		// progress.addDebug("Download chunks from " + downloadFrom +
		// " for URN " + urn + " / "
		// + getNodeID());
		final FutureGenericFile futureFileChunksOnePeer = new FutureGenericFileImpl();
		BitSet chunks = getChunkAv(downloadFrom, urn);
		if (chunks != null)
			chuknAV(urn, downloadFrom, progress, chunks, futureFileChunksOnePeer, psh);
		else
		{
			FutureResponse fr = fsRCP.chunkAvailability(downloadFrom, urn);
			// fr.reportDebugIn(progress);
			fr.addListener(new BaseFutureAdapter<FutureResponse>()
			{
				@Override
				public void operationComplete(FutureResponse future) throws Exception
				{
					// progress.addDebug("op complete");
					if (future.isFailed())
					{
						if (!noNeededParts.containsKey(downloadFrom))
							noNeededParts.put(downloadFrom, System.currentTimeMillis());
						logger.warn("future failed");
						futureFileChunksOnePeer
								.failed(Message.TYPE_NOT_OK, "get chunk info failed");
					}
					else
					{
						Collection<ByteArrayObject> bao = future.getResponseMessage().payloadAt(0);
						if (bao != null && bao.size() > 0)
						{
							BitSet chunks = (BitSet) bao.iterator().next().getObject();
							chuknAV(urn, downloadFrom, progress, chunks, futureFileChunksOnePeer,
									psh);
						}
						else
						{
							if (!noNeededParts.containsKey(downloadFrom))
								noNeededParts.put(downloadFrom, System.currentTimeMillis());
							// progress.addDebug("no needed parts3");
							logger.warn("no needed parts3");
							futureFileChunksOnePeer.failed(Message.TYPE_NOT_OK, "no needed parts3");
						}
					}
				}
			});
		}
		return futureFileChunksOnePeer;
	}

	public void removeURN(URN urn)
	{
		fileChunkMapping.removeMapping(urn.getHash());
	}

	private void chuknAV(final URN urn, final NodeAddress downloadFrom, final Progress progress,
			BitSet chunks, final FutureGenericFile futureFileChunksOnePeer, boolean psh)
	{
		putChunkAv(downloadFrom, urn, chunks);
		logger.debug("Available chunk size: " + chunks.cardinality() + " on " + downloadFrom
				+ " for " + urn + " / " + getNodeID());
		// progress.addDebug("Available chunk size: " + chunks.cardinality() +
		// " on " + downloadFrom
		// + " for " + urn + " / " + getNodeID());
		RandomBitSet rbs;
		if (fileChunkMapping.getFileChunks(urn) != null)
			rbs = fileChunkMapping.getFileChunks(urn).diff(chunks);
		else
			rbs = new RandomBitSet(chunks);
		logger.debug("Diff chunks: " + chunks.cardinality());
		// progress.addDebug("Diff chunks: " + chunks.cardinality());
		if (rbs != null && rbs.cardinality() > 0)
		{
			downloadChunks(urn, downloadFrom, progress, rbs, futureFileChunksOnePeer, psh);
		}
		else
		{
			if (!noNeededParts.containsKey(downloadFrom))
				noNeededParts.put(downloadFrom, System.currentTimeMillis());
			logger.warn("no needed parts1");
			history.removeTradeHint(downloadFrom);
			futureFileChunksOnePeer.failed(Message.TYPE_NOT_OK, "no needed parts1");
		}
	}

	private void downloadChunks(final URN urn, final NodeAddress downloadFrom,
			final Progress progress, final RandomBitSet rbs,
			final FutureGenericFile futureFileChunksOnePeer, final boolean psh)
	{
		FutureFileChunk ffc = downloadChunk(urn, rbs.getRandomChunkNr(), downloadFrom, progress,
				psh);
		Recursive.loop(new Loopable<FutureFileChunk>()
		{
			@Override
			public FutureFileChunk action() throws Exception
			{
				int chunkID = rbs.getRandomChunkNr();
				if (chunkID >= 0)
					return downloadChunk(urn, chunkID, downloadFrom, progress, psh);
				else
					return null;
			}

			@Override
			public void actionCompleted(FutureFileChunk future)
			{
				if (future.isSuccess() && fileChunkMapping.isComplete(future.getURN()))
				{
					history.removeTradeHint(downloadFrom);
					File complete = fileChunkMapping.getCompleteFile(future.getURN());
					logger.debug("This file is complete! " + future.getURN());
					futureFileChunksOnePeer.setFile(complete);
				}
				else if (future.isSuccess())
				{
					// progress.addDebug("Not complet " + future.getURN() +
					// " / " + getNodeID());
					logger.debug("Not complet " + future.getURN() + " / " + getNodeID());
					System.err.println("Not complet " + future.getURN() + " / "
							+ future.getChunkID() + " going for next one");
				}
				else
				{
					// noNeededParts.put(downloadFrom,
					// System.currentTimeMillis());
					// progress.addDebug("Download chunk failed from this peer");
					logger.warn("Download chunk failed from this peer " + future.getFailedReason());
					// we have further information we can use here...
					history.tradeHint(downloadFrom);
					if (future.getHint() instanceof byte[])
						history.tradeHint(downloadFrom, (byte[]) future.getHint());
					else if (future.getHint() instanceof Collection)
						history.tradeHint(downloadFrom, (Collection<BigInteger>) future.getHint());
					futureFileChunksOnePeer.failed(Message.TYPE_NOT_OK,
							"Download chunk failed from this peer");
				}
			}

			@Override
			public void end()
			{
				if (!futureFileChunksOnePeer.isCompleted())
				{
					// progress.addDebug("(E)Download chunk from this peer failed: "
					// + urn + " / "
					// + getNodeID());
					logger.warn("(E)Download chunk from this peer failed: " + urn + " / "
							+ getNodeID());
					futureFileChunksOnePeer
							.failed(Message.TYPE_NOT_OK,
									"(E)Download chunk from this peer failed: "
							+ urn + " / " + getNodeID());
				}
				else
				{
					// progress.addDebug("(E)Download chunk from this peer success="
					// + futureFileChunksOnePeer.isSuccess() + " : " + urn +
					// " / "
					// + getNodeID());
					logger.debug("(E)Download chunk from this peer success="
							+ futureFileChunksOnePeer.isSuccess() + " : " + urn + " / "
							+ getNodeID());
				}
			}

			@Override
			public void exceptionCaught(Throwable cause)
			{
				cause.printStackTrace();
				// progress.addDebug(cause.toString());
			}

			@Override
			public boolean isDone()
			{
				// System.err.println("isDone " +
				// queue.size());
				return rbs.noMoreChunks() || futureFileChunksOnePeer.isCompleted();
			}
		}, ffc, 1);
	}

	public FutureFileChunk downloadChunk(final URN urn, final Integer chunkID,
			final NodeAddress downloadFrom, final Progress progress, final boolean psh)
	{
		final FutureFileChunk futureFileChunk = new FutureFileChunkImpl(urn, progress);
		logger.debug("Download chunk nr " + chunkID + " from " + downloadFrom);
		System.err.println("Download chunk nr " + chunkID + " from " + downloadFrom);
		// progress.addDebug("Download chunk nr " + chunkID + " from " +
		// downloadFrom);
		FutureResponse fr = fsRCP.get(downloadFrom, urn.getHash(), chunkID, null);
		// fr.reportDebugIn(progress);
		fr.addListener(new BaseFutureAdapter<FutureResponse>()
		{
			@Override
			public void operationComplete(FutureResponse future)
			{
				// progress.addDebug("op1 done");
				if (future.isFailed())
					futureFileChunk.failed("fsRCP get failed");
				else
				{
					Message response = future.getResponseMessage();
					if (response == null)
						futureFileChunk.failed("response is null, this is weird");
					else
					{
						Collection<ByteArrayObject> tmp = response.payloadAt(0);
						if (tmp != null && response.getType() != Message.TYPE_DENIED)
						{
							int amount2 = addChunk(urn, tmp, chunkID);
							history.increaseDownload(downloadFrom, amount2);
							logger.debug("Downlead chunk successful " + response);
							futureFileChunk.chunkDownloaded(chunkID);
						}
						else if (tmp != null)
						{
							logger.debug("download denied " + response);
							// progress.addDebug("download denied " + response);
							//
							Object tmp2 = tmp.iterator().next().getObject();
							// if (currentStrategy == Strategy.PSH2)
							if (tmp2 instanceof byte[])
							{
								System.err.println("going PSH2");
								if (psh)
									psh2(tmp, downloadFrom, splitSize < urn.getSize() ? splitSize
											: urn.getSize(), futureFileChunk, urn, chunkID,
											progress);
								else
									futureFileChunk.failed("PSH2 failed (denied)", tmp2);
							}
							// else if (currentStrategy == Strategy.PSHR)
							else if (tmp2 instanceof Collection)
							{
								System.err.println("going PSHR");
								if (psh)
									pshr(tmp, downloadFrom, splitSize < urn.getSize() ? splitSize
											: urn.getSize(), futureFileChunk, urn, chunkID,
											progress);
								else
									futureFileChunk.failed("PSHR failed (denied)",
											tmp2);
							}
							else
							{
								// progress.addDebug("failed1");
								futureFileChunk.failed("This must be TFT?" + tmp);
							}
						}
						else
						{
							// denied.put(downloadFrom,
							// System.currentTimeMillis());
							logger.warn("Error downloading " + response + " /tmp " + tmp + " psh="
									+ psh);
							// progress.addDebug("failed2");
							futureFileChunk.failed("TFT failed (denied)");
						}
					}
				}
			}
		});
		return futureFileChunk;
	}

	private void pshr(Collection<ByteArrayObject> tmp, final NodeAddress downloadFrom, long amount,
			final FutureFileChunk futureFileChunk, final URN urn, final int chunkID,
			final Progress progress)
	{
		logger.info("downloaded failed, going for PSHR, " + amount);
		// progress.addDebug("downloaded failed, going for PSHR, " + amount);
		Collection<BigInteger> list = (Collection<BigInteger>) tmp.iterator().next().getObject();
		logger.debug("The node list size for PSH_r is " + list.size());
		if (list.size() == 0)
		{
			if (!noIntermediatePeers.containsKey(downloadFrom))
				noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
			futureFileChunk.failed("list size is zero");
			return;
		}
		Collection<NodeAddress> intermediateNodes = history.findPath(list, (int) amount,
				getNodeAddress(), downloadFrom);
		if (intermediateNodes.size() == 0)
		{
			if (!noIntermediatePeers.containsKey(downloadFrom))
				noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
			futureFileChunk.failed("there are no intermediate peers", list);
			return;
		}
		requestCheck(intermediateNodes, downloadFrom, amount, progress).addListener(
				new BaseFutureAdapter<FutureCheck>()
				{
					@Override
					public void operationComplete(FutureCheck future)
					{
						if (future.isSuccess())
						{
							Check check = future.getCheck();
							logger.debug("Rerequest with check " + check);
							// progress.addDebug("Rerequest with check " +
							// check);
							rerequest(downloadFrom, check, futureFileChunk, urn, chunkID, progress);
						}
						else
						{
							if (!noIntermediatePeers.containsKey(downloadFrom))
								noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
							logger.warn("No check for any intermediate peer found");
							// progress.addDebug("No check for any intermediate peer found");
							futureFileChunk.failed("No check for any intermediate peer found");
						}
					}
				});
	}

	private void psh2(Collection<ByteArrayObject> tmp, final NodeAddress downloadFrom, long amount,
			final FutureFileChunk futureFileChunk, final URN urn, final int chunkID,
			final Progress progress)
	{
		logger.info("downloaded failed, going for PSH2, " + amount);
		// progress.addDebug("downloaded failed, going for PSH2, " + amount);
		// BigInteger bloomFilter = (BigInteger)
		// tmp.iterator().next().getObject();
		byte[] bloomFilter = tmp.iterator().next().getArray();
		logger.debug("bloomfilter is " + bloomFilter + " / " + bloomFilter.length);
		if (bloomFilter.length == 0)
		{
			if (!noIntermediatePeers.containsKey(downloadFrom))
				noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
			// progress.addDebug("bloomfilter is 0 from " + downloadFrom);
			futureFileChunk.failed("bloomfilter is 0 from " + downloadFrom);
			System.err.println("no bf");
			return;
		}
		Collection<NodeAddress> intermediateNodes = history.findPath(bloomFilter, (int) amount,
				getNodeAddress(), downloadFrom);
		if (intermediateNodes.size() == 0)
		{
			if (!noIntermediatePeers.containsKey(downloadFrom))
				noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
			// progress.addDebug("no intermed. nodes");
			logger.warn("no intermed. nodes");
			futureFileChunk.failed("no intermed. nodes", bloomFilter);
			System.err.println("no inter");
			/*
			 * synchronized (PSH2.fileSharingNodes) { System.err.println("I'm "
			 * + getNodeAddress() + " dwn is " + downloadFrom); for (int i = 0;
			 * i < PSH2.fileSharingNodes.length; i++) { if
			 * (PSH2.fileSharingNodes[i].getNodeAddress().equals(downloadFrom))
			 * {
			 * PSH2.fileSharingNodes[i].history.dumpHistoryBeiDenenIchSchuldenHabe
			 * (amount); byte[] bf =
			 * PSH2.fileSharingNodes[i].history.bloomFilter((int) amount);
			 * System.err.println("BF is good!! " + Arrays.equals(bf,
			 * bloomFilter)); } }
			 * history.dumpHistoryDieBeiMirSchuldenHaben(amount); }
			 * intermediateNodes = history.findPath(bloomFilter, (int) amount,
			 * getNodeAddress(), downloadFrom); System.err.println("no inter " +
			 * intermediateNodes.size());
			 */
			return;
		}
		requestCheck(intermediateNodes, downloadFrom, amount, progress).addListener(
				new BaseFutureAdapter<FutureCheck>()
				{
					@Override
					public void operationComplete(FutureCheck future)
					{
						if (future.isSuccess())
						{
							Check check = future.getCheck();
							logger.info("Rerequest with check " + check);
							// progress.addDebug("Rerequest with check " +
							// check);
							rerequest(downloadFrom, check, futureFileChunk, urn, chunkID, progress);
						}
						else
						{
							if (!noIntermediatePeers.containsKey(downloadFrom))
								noIntermediatePeers.put(downloadFrom, System.currentTimeMillis());
							logger.warn("No check for any intermediate peer found");
							// progress.addDebug("No check for any intermediate peer found");
							futureFileChunk.failed("No check for any intermediate peer found");
						}
					}
				});
	}

	private FutureCheck requestCheck(Collection<NodeAddress> intermediateNodes,
			final NodeAddress target, final long amount, final Progress progress)
	{
		final FutureCheck futureCheck = new FutureCheckImpl();
		final Queue<NodeAddress> queue = new ConcurrentLinkedQueue<NodeAddress>();
		for (NodeAddress candidate : intermediateNodes)
		{
			if (!noIntermediatePeers.containsKey(candidate))
				queue.add(candidate);
		}
		if (queue.size() == 0)
		{
			futureCheck.failed(Message.TYPE_NOT_OK, "queue size is 0");
		}
		else
		{
			NodeAddress candidate = queue.poll();
			FutureResponse fr = fsRCP.requestBalance(candidate, target, (int) amount);
			Recursive.loop(new Loopable<FutureResponse>()
			{
				@Override
				public FutureResponse action() throws Exception
				{
					NodeAddress candidate = queue.poll();
					if (candidate != null)
					{
						// progress.addDebug("request");
						return fsRCP.requestBalance(candidate, target, (int) amount);
					}
					else
						return null;
				}

				@Override
				public void actionCompleted(FutureResponse future)
				{
					Message responseMessage = future.getResponseMessage();
					if (future.isSuccess() && responseMessage != null
							&& responseMessage.getType() == Message.TYPE_OK)
					{
						int amount2 = responseMessage.integerAt(0);
						logger.debug("PSH success!!! " + amount2);
						// progress.addDebug("PSH success!!! " + amount2);
						System.err.println("PSH success!!! " + amount2);
						futureCheck.setCheck(new Check(future.getResponseMessage().getSenderNode(),
								target, amount2));
					}
					else
					{
						if (!noIntermediatePeers.containsKey(future.getRequestMessage()
								.getRecipientNode()))
							noIntermediatePeers.put(future.getRequestMessage().getRecipientNode(),
									System.currentTimeMillis());
						logger.debug("PSH failure!!! + responseMessage");
						// progress.addDebug("PSH failure!!! ");
						System.err.println("PSH failure!!! (" + System.currentTimeMillis() + ") "
								+ responseMessage);
					}
				}

				@Override
				public void end()
				{
					if (!futureCheck.isCompleted())
					{
						logger.warn("Request check failed.");
						// progress.addDebug("Request check failed.");
						futureCheck.failed(Message.TYPE_NOT_OK, "Request check failed.");
					}
				}

				@Override
				public void exceptionCaught(Throwable cause)
				{
					cause.printStackTrace();
					// progress.addDebug(cause.toString());
				}

				@Override
				public boolean isDone()
				{
					return queue.size() == 0 || futureCheck.isCompleted();
				}
			}, fr, PARALLEL);
		}
		return futureCheck;
	}

	private void rerequest(final NodeAddress downloadFrom, final Check check,
			final FutureFileChunk futureFileChunk, final URN urn, final int chunkID,
			final Progress progress)
	{
		history.applyCheckSender(downloadFrom, check);
		fsRCP.get(downloadFrom, urn.getHash(), chunkID, check).addListener(
				new BaseFutureAdapter<FutureResponse>()
				{
					@Override
					public void operationComplete(FutureResponse future)
					{
						if (future.isFailed())
						{
							unapply(check);
							// progress.addDebug("FF1");
							futureFileChunk.failed("FF1");
							System.err.println("FF1");
							return;
						}
						Message response = future.getResponseMessage();
						if (response == null)
						{
							unapply(check);
							// progress.addDebug("FF2");
							futureFileChunk.failed("FF2");
							System.err.println("FF2");
							return;
						}
						if (response.getType() == Message.TYPE_DENIED)
						{
							unapply(check);
							// progress.addDebug("FF3");
							futureFileChunk.failed("FF3");
							System.err.println("FF3");
							return;
						}
						Collection<ByteArrayObject> tmp = response.payloadAt(0);
						if (tmp == null)
						{
							unapply(check);
							// progress.addDebug("FF4");
							futureFileChunk.failed("FF4");
							System.err.println("FF4");
							return;
						}
						else if (tmp.size() > 0)
						{
							int amout2 = addChunk(urn, tmp, chunkID);
							history.increaseDownload(downloadFrom, amout2);
							// progress.addDebug("OK");
							futureFileChunk.chunkDownloaded(chunkID);
						}
						else
						{
							unapply(check);
							// progress.addDebug("FF5");
							futureFileChunk.failed("FF5");
							System.err.println("FF5");
							return;
						}
					}
				});
	}

	private void unapply(Check check)
	{
		// now we lost this resources in the check, check.getIntermediat peer is
		// not nice to us!!
		System.err.println("LOST MONEY!! This should rarely happen in simulation");
		// target thinks that he can download in intermediate, but the amount
		// has been transferred to source
	}

	private URN getURN(File f) throws URISyntaxException
	{
		if (files == null || urns == null)
			return new URN(f);
		else
		{
			int index = files.indexOf(f.getName());
			return urns.get(index);
		}
	}

	private void initFilesInPublishDir() throws IOException
	{
		File[] fs = publishDir.listFiles(new FileFilter()
		{
			@Override
			public boolean accept(File pathname)
			{
				return pathname.isFile();
			}
		});
		for (File f : fs)
		{
			try
			{
				addFileAndPublish(f, getURN(f));
			}
			catch (URISyntaxException e)
			{
				e.printStackTrace();
				System.err.println("cannot publish this file...");
			}
		}
	}

	public void addFile(File file) throws IOException, URISyntaxException
	{
		addFile(file, new URN(file));
	}

	public void addFile(File file, URN urn) throws IOException, URISyntaxException
	{
		logger.debug("Add file in my local repo, I'm " + getNodeID());
		File dest = new File(publishDir, file.getName());
		if (!dest.exists())
			FileUtils.copyFile(file, dest);
	}

	public Strategy getCurrentStrategy()
	{
		return currentStrategy;
	}

	public void setFreeriding(boolean freeriding)
	{
		this.freeriding = freeriding;
	}

	public boolean isFreerider()
	{
		return freeriding;
	}

	public int addChunk(URN urn, Collection<ByteArrayObject> tmp, int chunkID)
	{
		ByteArrayObject bao = tmp.iterator().next();
		byte[] me = bao.getArray();
		int length = me.length;
		logger.debug("Publish chunk  " + chunkID + " with file length " + me.length + " hash:"
				+ Utils.makeSHAHash(me));
		System.err.println("Publish chunk  " + chunkID + " with file length " + me.length
				+ " hash:" + Utils.makeSHAHash(me));
		try
		{
			ByteBuffer bb = ByteBuffer.wrap(me);
			fileChunkMapping.addChunk(urn, chunkID, bb);
			bb = null;
			me = null;
			if (publishedURN.add(urn))
				publishURN(urn);
		}
		catch (IOException e)
		{
			// TODO: handle if disk is fully or other issues
			e.printStackTrace();
		}
		return length;
	}

	public void setListenDate(int port, int port2, BindInformation bindInformation,
			Replication replication, File dir, InetSocketAddress boot)
	{
		listenData.setPortUPD(port);
		listenData.setPortTPC(port2);
		listenData.setBindInformation(bindInformation);
		listenData.setReplication(replication);
		listenData.setDir(dir);
		listenData.setBoot(boot);
	}

	public InetSocketAddress listen() throws IOException
	{
		listen(listenData.getPortUDP(), listenData.getPortTCP(), listenData.getBindInformation(),
				listenData.getReplication(), listenData.getDir());
		return listenData.getBoot();
	}

	public void afterBootstrap() throws IOException
	{
		initFilesInPublishDir();
	}

	private void afterListen() throws IOException
	{
		System.err.println("After Listen");
		getStore().addFirstFilter(new ReplyFilter()
		{
			@Override
			public void handleResponse(ReplyFilter nextFilter, IoSession session, Message message,
					StoreMultiMap source) throws IOException
			{
				try
				{
					String domain = message.stringAt(0);
					if ((message.getName() == Message.NAME_ADD || message.getName() == Message.NAME_PUT)
							&& message.getType() == Message.TYPE_REQUEST)
					{
						// add sender node, we do not need to send this info
						// twice over the wire
						if (TRACKERS.equals(domain))
						{
							ArrayList<NodeAddress> tmp = new ArrayList<NodeAddress>(1);
							tmp.add(message.getSenderNode());
							message.setNeighborsAt(tmp, 0);
						}
					}
					if (message.keyAt(0) == null)
					{
						logger.error("wrong (null) key in message" + message);
					}
					nextFilter.handleResponse(nextFilter.getNextFilter(), session, message, source);
				}
				catch (Exception e)
				{
					e.printStackTrace();
				}
			}
		});
		fileChunkMapping = new FileChunkMapping(incomingDir, splitSize);
		fsRCP = new FileSharingRPC(this, getNodeInformation(), history, fileChunkMapping,
				currentStrategy, splitSize, goodwill);
	}

	public FileChunkMapping getFileChunkMapping()
	{
		return fileChunkMapping;
	}
}

class TimeBitSet extends BitSet
{
	private final long time;

	public TimeBitSet(BitSet data)
	{
		this.time = System.currentTimeMillis();
		or(data);
	}

	public long getTime()
	{
		return time;
	}
}