package ch.uzh.csg.filesharing;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.BitSet;
import java.util.Collection;
import java.util.Random;

import org.apache.mina.core.future.IoFuture;
import org.apache.mina.core.future.IoFutureListener;
import org.apache.mina.core.future.WriteFuture;
import org.apache.mina.core.session.IoSession;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.uzh.csg.filesharing.FileSharingNode.Strategy;
import ch.uzh.csg.filesharing.fileutils.FileChunkMapping;
import ch.uzh.csg.filesharing.fileutils.SplitterUtils;
import ch.uzh.csg.tomp2p.Configuration;
import ch.uzh.csg.tomp2p.Utils;
import ch.uzh.csg.tomp2p.common.FutureResponse;
import ch.uzh.csg.tomp2p.connection.NodeInformationBean;
import ch.uzh.csg.tomp2p.message.Message;
import ch.uzh.csg.tomp2p.routing.NodeAddress;
import ch.uzh.csg.tomp2p.rpc.ReplyHandler;
import ch.uzh.csg.tomp2p.storage.ByteArrayObject;

public class FileSharingRPC extends ReplyHandler
{
	final private static Logger logger = LoggerFactory.getLogger(FileSharingRPC.class);
	static int successDirect = 0;
	static int failedDirect = 0;
	static int successInDirect = 0;
	static int failedInDirect = 0;
	final private History history;
	final private FileChunkMapping fileChunkMapping;
	final private static byte GET_CHECK = Message.NAME_USER_DEFINED1;
	final private static byte CHUNK_AVAILIBILITY = Message.NAME_USER_DEFINED2;
	final private static byte FS_GET = Message.NAME_USER_DEFINED3;
	// final public static byte NO_SLOT = Message.Type.USER_DEFINED2;
	final private Strategy currentStrategy;
	final private int splitSize;
	final private double goodwill;

	public FileSharingRPC(FileSharingNode fileSharingNode, NodeInformationBean nodeInformationBean,
			History history, FileChunkMapping fileChunkMapping, Strategy currentStrategy,
			int splitSize, double goodwill)
	{
		super(fileSharingNode.getP2PID(), nodeInformationBean);
		registerIoHandler(GET_CHECK, CHUNK_AVAILIBILITY, FS_GET);
		this.history = history;
		this.fileChunkMapping = fileChunkMapping;
		this.currentStrategy = currentStrategy;
		this.splitSize = splitSize;
		this.goodwill = goodwill;
	}

	public FutureResponse requestBalance(NodeAddress nodeToAsk, final NodeAddress target, int amount)
	{
		logger.debug("Send requestBalance to " + nodeToAsk);
		Message message = createMessage(nodeToAsk, GET_CHECK, Message.TYPE_REQUEST);
		message.payload(Utils.wrapCollection(target));
		message.integer(amount);
		return getRequest().send(message);
	}

	public FutureResponse get(NodeAddress downloadFrom, BigInteger hash, int chunkID, Check check)
	{
		logger.debug("Send get (fsRPC) to " + downloadFrom);
		Message message = createMessage(downloadFrom, FS_GET, Message.TYPE_REQUEST);
		message.key(hash);
		message.integer(chunkID);
		if (check != null)
			message.payload(Utils.wrapCollection(check));
		return getRequest().send(message, splitSize > Configuration.UDP_LENGTH);
	}

	public FutureResponse chunkAvailability(NodeAddress downloadFrom, URN urn)
	{
		logger.debug("Send chunkDiff (fsRPC) to " + downloadFrom);
		Message message = createMessage(downloadFrom, CHUNK_AVAILIBILITY, Message.TYPE_REQUEST);
		message.key(urn.getHash());
		return getRequest().send(message,
				SplitterUtils.numPieces(urn.getSize(), splitSize) > 1000 * 8);
	}

	@Override
	public void handleResponse(IoSession session, Message message) throws IOException
	{
		logger.debug("handleResponse " + message.getType());
		if (message.getType() == Message.TYPE_REQUEST)
		{
			// switch statement does not work here
			if (message.getName() == GET_CHECK)
				handleGetCheck(session, message);
			else if (message.getName() == FS_GET)
				handleFSGet(session, message);
			else if (message.getName() == CHUNK_AVAILIBILITY)
				handleChunk(session, message);
		}
	}

	private void handleChunk(IoSession session, Message message) throws IOException
	{
		BigInteger key = message.keyAt(0);
		BitSet chunks = fileChunkMapping.getChunks(key);
		Message responseMessage;
		if (chunks == null)
			responseMessage = createMessage(message.getSenderNode(), CHUNK_AVAILIBILITY,
					Message.TYPE_NOT_FOUND);
		else
		{
			responseMessage = createMessage(message.getSenderNode(), CHUNK_AVAILIBILITY,
					Message.TYPE_OK);
			responseMessage.payload(Utils.wrapCollection(chunks));
		}
		response(session, message, responseMessage);
	}
	final private static Random rnd = new Random();

	private void handleFSGet(IoSession session, final Message message) throws IOException
	{
		BigInteger key = message.keyAt(0);
		int chunkID = message.integerAt(0);
		Collection<ByteArrayObject> tmp = message.payloadAt(0);
		if (tmp != null)
		{
			Check check = (Check) tmp.iterator().next().getObject();
			history.applyCheckRecipient(message.getSenderNode(), check);
			logger.debug("check applied from " + message.getSenderNode());
		}
		int timeout = session.getTransportMetadata().isConnectionless() ? Configuration
				.getTimeoutUDP() : Configuration.getTimeoutTCP();
		if (System.currentTimeMillis() - message.getStamp() > timeout)
		{
			session.close(false);
			return;
		}
		ByteBuffer buffer = fileChunkMapping.getChunk(key, chunkID);
		Message responseMessage;
		if (buffer != null)
		{
			final int length = buffer.limit();
			final NodeAddress sender = message.getSenderNode();
			boolean canDownloadGoodWill = rnd.nextDouble() < goodwill;
			boolean canDownloadCredit = history.canUploadTo(sender, length);
			// boolean isKnown = history.isKnown(sender);
			boolean isPotentialTrader = history.isPotentialTrader(sender);
			if (currentStrategy == Strategy.NONE
					|| ((isPotentialTrader && canDownloadCredit) || canDownloadGoodWill))
			{
				logger.debug("download possible for " + sender + " for chunk " + chunkID);
				System.err.println("download possible for " + sender + " for chunk " + chunkID
						+ " hash:" + Utils.makeSHAHash(buffer));
				buffer.flip();
				if (isPotentialTrader)
					System.err.println("Potential Trader");
				if (canDownloadGoodWill)
					System.err.println("Goodwill");
				responseMessage = createMessage(sender, FS_GET, Message.TYPE_OK);
				Collection<ByteArrayObject> value = Utils.wrapCollection(buffer);
				// Collection<ByteArrayObject> value = Utils.wrapCollection(me);
				responseMessage.payload(value);
				// we need to account here and rollback in case of errors.
				// Otherwise we migth grant more resources than credit.
				history.increaseDownload(sender, -length);
				IoFuture f = response(session, message, responseMessage);
				if (f instanceof WriteFuture)
				{
					f.addListener(new IoFutureListener<WriteFuture>()
					{
						@Override
						public void operationComplete(WriteFuture wf)
						{
							// rollback if failed
							if (!wf.isWritten())
								history.increaseDownload(sender, length);
						}
					});
					return;
				}
			}
			else
			{
				logger.debug("download check ok for " + sender);
				responseMessage = createDenied(message, length, Message.TYPE_DENIED);
			}
		}
		else
		{
			responseMessage = createMessage(message.getSenderNode(), FS_GET, Message.TYPE_NOT_FOUND);
		}
		response(session, message, responseMessage);
	}

	private Message createDenied(Message message, int length, byte type)
	{
		Message responseMessage = createMessage(message.getSenderNode(), FS_GET, type);
		if (currentStrategy == Strategy.PSH2)
		{
			byte[] bloomfilter = history.bloomFilter(length);
			logger.debug("cannot upload to, but exchange bloomfilter" + message.getSenderNode()
					+ " / " + history.toString(message.getSenderNode()) + " / length:"
					+ bloomfilter.length);
			responseMessage.payload(Utils.wrapCollection(bloomfilter));
		}
		else if (currentStrategy == Strategy.PSHR)
		{
			logger.debug("cannot upload to, but exchange data" + message.getSenderNode() + " / "
					+ history.toString(message.getSenderNode()));
			Collection<BigInteger> latestPeers = history.getLatestPeers(length);
			responseMessage.payload(Utils.wrapCollection(latestPeers));
		}
		return responseMessage;
	}

	private void handleGetCheck(IoSession session, Message message) throws IOException
	{
		logger.debug("handleGetCheck");
		NodeAddress na = (NodeAddress) message.payloadAt(0).iterator().next().getObject();
		// NodeAddress na = new NodeAddress(nId);
		int amount = message.integerAt(0);
		// remove download from sender, add to na
		// History.AccountingStatus status =
		// history.account(message.getSenderNode(), na, amount);
		int namount = history.account(message.getSenderNode(), na, amount);
		Message responseMessage;
		// if (status == AccountingStatus.OK)
		if (namount > 0)
		{
			responseMessage = createMessage(message.getSenderNode(), GET_CHECK, Message.TYPE_OK);
			// responseMessage.integer(amount);
			responseMessage.integer(namount);
			successInDirect++;
		}
		else
		{
			failedInDirect++;
			responseMessage = createMessage(message.getSenderNode(), GET_CHECK, Message.TYPE_DENIED);
			// responseMessage.integer(status.ordinal());
			responseMessage.integer(namount);
		}
		response(session, message, responseMessage);
		logger.debug("Bond Ratio=" + successInDirect + "/" + failedInDirect);
	}
	/*
	 * private int account(NodeAddress source, NodeAddress target) { Integer
	 * transferSource = history.get(source); Integer transferTarget =
	 * history.get(target); if (transferSource != null && transferTarget !=
	 * null) { if (transferSource < 0 && transferTarget > 0) { int balance =
	 * Math.min(Math.abs(transferSource), Math.abs(transferTarget));
	 * history.account(source, transferSource + balance);
	 * history.account(target, transferTarget - balance); //
	 * resource.overwriteHistory(source, transferSource + balance); //
	 * resource.overwriteHistory(target, transferTarget - balance); return
	 * balance; } } return 0; }
	 */
}