package ch.uzh.csg.filesharing;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import ch.uzh.csg.tomp2p.routing.NodeAddress;
import ch.uzh.csg.tomp2p.routing.Routing;

public class History
{
	final private int credit;
	final private int chunkSize;
	final private int timeToTreade;
	final private Map<NodeAddress, Amount> accounting = new Cache<NodeAddress, Amount>(1000);
	final private Map<NodeAddress, Set<NodeAddress>> tradeHints = new Cache<NodeAddress, Set<NodeAddress>>(
			100);
	final private static Logger logger = LoggerFactory.getLogger(History.class);
	final public static int MAX_PSHR_ENTRIES = 4;
	// payload is 499 bytes with 5 entries
	final private static int MAX_BLOOM = Routing.BITS * 25;
	final private static int MAX_BLOOM_ENTRIES = 100;
	// expectedFalsePositiveProbability = 0.677220983613415%
	// final private Collection<NodeAddress> waiting = new
	// HashSet<NodeAddress>();
	private byte[] bloomfilter = new byte[0];
	private long bloomfilterCreated = 0;
	private boolean changeBloomFilter = false;
	public enum AccountingStatus
	{
		OK, SOURCE_FAIL, TARGET_FAIL, BOTH_FAIL
	}

	public History(int chunkSize, int credit, int timeToTreade)
	{
		this.credit = credit;
		this.chunkSize = chunkSize;
		this.timeToTreade = timeToTreade;
	}

	private Amount get(NodeAddress key)
	{
		return accounting.get(key);
	}

	public int getAmountForTrade(NodeAddress key)
	{
		synchronized (accounting)
		{
			int amount = 0;
			Amount a = accounting.get(key);
			if (a != null && a.readyToTrade())
			{
				amount = a.getAmount();
			}
			return amount;
		}
	}

	public void applyCheckRecipient(NodeAddress sender, Check check)
	{
		synchronized (accounting)
		{
			increaseDownload(sender, check.getAmount());
			increaseDownload(check.getIntermediate(), -check.getAmount());
		}
	}

	public void applyCheckSender(NodeAddress recipient, Check check)
	{
		synchronized (accounting)
		{
			increaseDownload(recipient, -check.getAmount());
			increaseDownload(check.getIntermediate(), check.getAmount());
		}
	}

	public boolean canDownloadFrom(NodeAddress key, int length)
	{
		synchronized (accounting)
		{
			Amount amount = get(key);
			int downloaded = 0;
			if (amount != null)
				downloaded = amount.getAmount();
			logger.debug("canDownloadFrom " + key + ": " + (downloaded + length)
					+ "(downloaded + length) <=" + credit + " (credit)");
			return (downloaded + length) <= credit;
		}
	}

	public boolean canUploadTo(NodeAddress key, int length)
	{
		synchronized (accounting)
		{
			Amount amount = get(key);
			int downloaded = 0;
			if (amount != null)
				downloaded = amount.getAmount();
			logger.debug("canUploadTo " + key + ": " + (downloaded - length)
					+ " (downloaded - length)>=" + (-credit) + " (-credit)");
			return (downloaded - length) >= -credit;
		}
	}

	public boolean isKnown(NodeAddress key)
	{
		synchronized (accounting)
		{
			return accounting.containsKey(key);
		}
	}

	public void notifyChange()
	{
		synchronized (accounting)
		{
			accounting.notifyAll();
		}
	}

	public void waitForChange(int time)
	{
		synchronized (accounting)
		{
			try
			{
				accounting.wait(time);
			}
			catch (InterruptedException e)
			{
			}
		}
	}

	public void increaseDownload(NodeAddress key, int diff)
	{
		synchronized (accounting)
		{
			Amount amount = get(key);
			if (amount == null)
			{
				amount = new Amount(timeToTreade);
				accounting.put(key, amount);
			}
			amount.update();
			amount.addAmount(diff);
			logger
					.debug("Node:" + key + ", set new balance (" + diff + "):"
							+ (amount.getAmount()));
			if (amount.getAmount() <= chunkSize)
			{
				accounting.notifyAll();
			}
			changeBloomFilter = true;
		}
	}

	public byte[] bloomFilter()
	{
		return bloomFilter(chunkSize);
	}

	public byte[] bloomFilter(int length)
	{
		if (changeBloomFilter)
			// && System.currentTimeMillis() - bloomfilterCreated >
			// (timeToTreade / 10))
			bloomfilter = makeBloomFilter(length);
		return bloomfilter;
	}

	private byte[] makeBloomFilter(int requestSize)
	{
		synchronized (accounting)
		{
			SimpleBloomFilter<BigInteger> sbf = new SimpleBloomFilter<BigInteger>(MAX_BLOOM,
					MAX_BLOOM_ENTRIES);
			// System.err.println(sbf.expectedFalsePositiveProbability() * 100 +
			// "%");
			int counter = 0;
			for (NodeAddress key : accounting.keySet())
			{
				// bei dene ich Schulden habe
				if (getAmountForTrade(key) >= requestSize)
				{
					// System.err.println("ADD " + key);
					sbf.add(key.getID());
					counter++;
					if (counter > MAX_BLOOM_ENTRIES)
					{
						// System.err.println("BF reached max");
						break;
					}
				}
				// else
				// System.err.println("NOT ADD " + key + " / " +
				// accounting.get(key));
			}
			// System.err.println("expectedFalsePositiveProbability="
			// + sbf.expectedFalsePositiveProbability());
			// System.err.println("BF CREATE " + PSHUtils.toBigInteger(bs));
			changeBloomFilter = false;
			bloomfilterCreated = System.currentTimeMillis();
			return sbf.toByteArray();
		}
	}

	public Collection<BigInteger> getLatestPeers(int requestSize)
	{
		Collection<BigInteger> tmp = new HashSet<BigInteger>();
		int counter = 0;
		synchronized (accounting)
		{
			for (NodeAddress key : accounting.keySet())
			{
				if (getAmountForTrade(key) >= requestSize)
				{
					tmp.add(key.getID());
					counter++;
					if (counter > MAX_PSHR_ENTRIES)
						break;
				}
			}
			return tmp;
		}
	}
	class Cache<K, V> extends LinkedHashMap<K, V>
	{
		private final int maxEntries;

		Cache(int maxEntries)
		{
			this.maxEntries = maxEntries;
		}

		protected boolean removeEldestEntry(Map.Entry<K, V> eldest)
		{
			return size() > maxEntries;
		}
	}

	public Collection<NodeAddress> findPath(byte[] bf, int requestSize)
	{
		return findPath(bf, requestSize, new NodeAddress[0]);
	}

	public Collection<NodeAddress> findPath(byte[] bf, int requestSize, NodeAddress... exclude)
	{
		// System.err.println("BF AFTER " + bf);
		SimpleBloomFilter<BigInteger> sbf = new SimpleBloomFilter<BigInteger>(MAX_BLOOM,
				MAX_BLOOM_ENTRIES, bf);
		// System.err.println("false positive probability: " +
		// sbf.expectedFalsePositiveProbability());
		Collection<NodeAddress> retVal = new ArrayList<NodeAddress>();
		synchronized (accounting)
		{
			List<NodeAddress> tmp = Arrays.asList(exclude);
			for (NodeAddress candidate : accounting.keySet())
			{
				if (getAmountForTrade(candidate) <= -requestSize && !tmp.contains(candidate))
				{
					if (sbf.contains(candidate.getID()))
					{
						retVal.add(candidate);
					}
				}
			}
		}
		return retVal;
	}

	public Collection<NodeAddress> findPath(Collection<BigInteger> list, int requestSize,
			NodeAddress... exclude)
	{
		Collection<NodeAddress> retVal = new ArrayList<NodeAddress>();
		synchronized (accounting)
		{
			List<NodeAddress> tmp = Arrays.asList(exclude);
			for (NodeAddress candidate : accounting.keySet())
			{
				if (getAmountForTrade(candidate) <= -requestSize && !tmp.contains(candidate))
				{
					if (list.contains(candidate.getID()))
					{
						retVal.add(candidate);
					}
				}
			}
		}
		return retVal;
	}

	public AccountingStatus account2(NodeAddress source, NodeAddress target, int amount)
	{
		synchronized (accounting)
		{
			if (isSourceOK(source, amount) && isTragetOK(target, amount))
			{
				{
					increaseDownload(source, -amount);
					increaseDownload(target, amount);
					return AccountingStatus.OK;
				}
			}
			else if (isSourceOK(source, amount))
				return AccountingStatus.TARGET_FAIL;
			else if (isTragetOK(target, amount))
				return AccountingStatus.SOURCE_FAIL;
			else
			{
				System.err.println("transferSource failed");
				return AccountingStatus.BOTH_FAIL;
			}
		}
	}

	public int account(NodeAddress source, NodeAddress target, int amount)
	{
		synchronized (accounting)
		{
			// max
			int tmps = getAmountForTrade(source);
			int tmpt = getAmountForTrade(target);
			amount = Math.min(Math.abs(tmps), Math.abs(tmpt));
			if (isSourceOK(source, amount) && isTragetOK(target, amount) && amount > 0)
			{
				{
					increaseDownload(source, -amount);
					increaseDownload(target, amount);
					return amount;
				}
			}
			else if (amount == 0)
				return 0;
			else if (isSourceOK(source, amount))
				// AccountingStatus.TARGET_FAIL;
				return -1;
			// AccountingStatus.SOURCE_FAIL;
			else if (isTragetOK(target, amount))
				return -2;
			// AccountingStatus.BOTH_FAIL;
			else
				return -3;
		}
	}

	private boolean isSourceOK(NodeAddress source, int amount)
	{
		// make sure that accounting is synced outside
		Amount transferSource = get(source);
		return transferSource != null && transferSource.getAmount() - amount >= 0;
	}

	private boolean isTragetOK(NodeAddress target, int amount)
	{
		// make sure that accounting is synced outside
		Amount transferTarget = get(target);
		return transferTarget != null && transferTarget.getAmount() + amount <= 0;
	}

	public void removeAccounting(NodeAddress key)
	{
		synchronized (accounting)
		{
			accounting.remove(key);
		}
	}

	public int assetSize(int min)
	{
		int counter = 0;
		synchronized (accounting)
		{
			for (NodeAddress key : accounting.keySet())
			{
				// bei dene ich Schulden habe
				if (getAmountForTrade(key) >= min)
					counter++;
			}
		}
		return counter;
	}

	public boolean isPotentialTrader(NodeAddress downloadFrom)
	{
		synchronized (accounting)
		{
			return tradeHints.containsKey(downloadFrom);
		}
	}

	public void tradeHint(NodeAddress firstHand)
	{
		tradeHint(firstHand, firstHand);
	}

	public void tradeHint(NodeAddress firstHand, NodeAddress secondHand)
	{
		synchronized (accounting)
		{
			// we want to have the latest entry first, so remove and insert the
			// item
			Set<NodeAddress> tmp = tradeHints.get(secondHand);
			if (tmp == null)
			{
				tmp = new HashSet<NodeAddress>();
				tradeHints.put(secondHand, tmp);
			}
			tmp.add(firstHand);
		}
	}

	public void removeTradeHint(NodeAddress firstHand)
	{
		synchronized (accounting)
		{
			for (Iterator<NodeAddress> i = tradeHints.keySet().iterator(); i.hasNext();)
			{
				NodeAddress secondHand = i.next();
				Set<NodeAddress> tmp = tradeHints.get(secondHand);
				if (tmp != null)
				{
					tmp.remove(firstHand);
					if (tmp.size() == 0)
						i.remove();
				}
			}
		}
	}

	public void tradeHint(NodeAddress firstHand, byte[] hint)
	{
		SimpleBloomFilter<BigInteger> sbf = new SimpleBloomFilter<BigInteger>(MAX_BLOOM,
				MAX_BLOOM_ENTRIES, hint);
		synchronized (accounting)
		{
			for (NodeAddress candidate : accounting.keySet())
			{
				if (sbf.contains(candidate.getID()))
					tradeHint(firstHand, candidate);
			}
		}
	}

	public void tradeHint(NodeAddress firstHand, Collection<BigInteger> hint)
	{
		synchronized (accounting)
		{
			for (NodeAddress candidate : accounting.keySet())
			{
				if (hint.contains(candidate.getID()))
					tradeHint(firstHand, candidate);
			}
		}
	}

	public String toString(NodeAddress key)
	{
		StringBuilder sb = new StringBuilder("History balance for ");
		sb.append(key).append("=");
		synchronized (accounting)
		{
			sb.append(get(key));
		}
		return sb.toString();
	}
}

class Amount
{
	final private int timeToTreade;
	private int amount;
	private long stamp = Long.MAX_VALUE;

	Amount(int timeToTreade)
	{
		this.timeToTreade = timeToTreade;
	}

	public void update()
	{
		stamp = System.currentTimeMillis();
	}

	boolean readyToTrade()
	{
		// only trade amounts that are old!
		return System.currentTimeMillis() - stamp >= timeToTreade;
	}

	public void addAmount(int amount)
	{
		this.amount += amount;
		update();
	}

	public int getAmount()
	{
		return amount;
	}

	@Override
	public String toString()
	{
		return "Amount=" + getAmount();
	}
}