package ch.uzh.csg.filesharing;
import java.io.File;
import java.io.Serializable;
import java.math.BigInteger;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.StringTokenizer;

import ch.uzh.csg.tomp2p.Utils;

/**
 * This class represents the naming scheme for this P2P file sharing
 * application. It is used as follows: afs:name!size!hash. This is complient
 * with RFC 2396 using the "opaque_part = uric_no_slash *uric" notation. The
 * hash is encoded in base64.
 * 
 * 
 * @author draft
 */
public class URN implements Serializable
{
	public static final long serialVersionUID = 5602341776850396834L;
	public final static String SCHEME = "afs";
	private String name;
	private BigInteger hash;
	private long size;
	// we cannot use /, +, or = as those are used by base64. We use !, which is
	// an unreserved character
	protected final static String DELIM = "!";

	public URN(URI uri) throws URISyntaxException
	{
		if (uri == null)
			throw new URISyntaxException("null", "URI is null");
		if (!SCHEME.equals(uri.getScheme()))
			throw new URISyntaxException(uri.toString(), "Can only deal with " + SCHEME);
		parse(uri.getSchemeSpecificPart());
	}

	public URN(File file) throws URISyntaxException
	{
		setName(file.getName());
		setHash(Utils.makeSHAHash(file));
		setSize(file.length());
	}

	public URN(String name, long size, BigInteger hash) throws URISyntaxException
	{
		setName(name);
		setHash(hash);
		setSize(size);
	}

	private void parse(String ssp) throws URISyntaxException
	{
		StringTokenizer st = new StringTokenizer(ssp, DELIM);
		if (st.countTokens() < 3)
			throw new RuntimeException("URN needs to be of format: afs:name|size|hash");
		setName(st.nextToken());
		try
		{
			setSize(Long.parseLong(st.nextToken()));
			setHash(st.nextToken());
		}
		catch (NumberFormatException nfe)
		{
			throw new RuntimeException("URN needs to be of format: afs:name|size|hash - " + nfe);
		}
		while (st.hasMoreTokens())
		{
			extensions(st.nextToken());
		}
	}

	protected void extensions(String nextToken)
	{
		// for future use
	}

	public void setName(String name) throws URISyntaxException
	{
		if (name.indexOf(DELIM) >= 0)
			throw new URISyntaxException(name, "The name cannot contain [" + DELIM + "]");
		this.name = name;
	}

	public String getName()
	{
		return name;
	}

	private void setHash(String nextToken)
	{
		byte[] data = ConversionUtils.decode(nextToken.toCharArray());
		this.hash = new BigInteger(1, data);
	}

	private void setHash(BigInteger hash)
	{
		this.hash = hash;
	}

	public BigInteger getHash()
	{
		return hash;
	}

	private void setSize(long size)
	{
		this.size = size;
	}

	public long getSize()
	{
		return size;
	}

	public String toString()
	{
		StringBuilder sb = new StringBuilder("afs:");
		sb.append(getName());
		sb.append(DELIM);
		sb.append(getSize());
		sb.append(DELIM);
		byte[] data = getHash().toByteArray();
		sb.append(ConversionUtils.encode(data));
		return sb.toString();
	}

	@Override
	public boolean equals(Object obj)
	{
		if (obj instanceof URN)
		{
			URN urn = (URN) obj;
			// we do not check the name, as the hash tells us wheter its the
			// same
			return hash.equals(urn.hash) && size == urn.size;
		}
		else
			return false;
	}

	@Override
	public int hashCode()
	{
		return (int) (size ^ (size >>> 32)) ^ (hash.hashCode());
	}
}
