package ch.uzh.csg.filesharing.fastss;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;

import ch.uzh.csg.filesharing.FileSharingNode;
import ch.uzh.csg.filesharing.URN;
import ch.uzh.csg.filesharing.FileSharingNode.Strategy;
import ch.uzh.csg.filesharing.future.FutureFile;
import ch.uzh.csg.filesharing.future.FuturePublish;
import ch.uzh.csg.tomp2p.Replication;
import ch.uzh.csg.tomp2p.Utils;
import ch.uzh.csg.tomp2p.action.FutureDHT;
import ch.uzh.csg.tomp2p.common.BaseFuture;
import ch.uzh.csg.tomp2p.common.BaseFutureAdapter;
import ch.uzh.csg.tomp2p.connection.BindInformation;
import ch.uzh.csg.tomp2p.storage.MultiMapMemory;

public class FileSharingFastSS
{
	// TAG for Debug-Output-messages
	public static final int P2PID = 44;
	private final static String domain = "FastSS";
	final private FileSharingNode fsn;

	public FileSharingFastSS(BigInteger id, int udpPort, int tcpPort,
			BindInformation bindInformation, Replication replication, File homeDir)
			throws IOException, URISyntaxException
	{
		this(id, udpPort, tcpPort, bindInformation, replication, new File(homeDir, "published"),
				new File(homeDir, "incoming"), new File(homeDir, "data"));
	}

	public FileSharingFastSS(BigInteger id, int udpPort, int tcpPort,
			BindInformation bindInformation, Replication replication, File publishDir,
			File incomingDir, File dataDir) throws IOException, URISyntaxException
	{
		// BigInteger id = Utils.createRandomNodeID();
		fsn = new FileSharingNode(P2PID, id, Strategy.NONE, 0, 1300, null, null, publishDir,
				incomingDir, 0, 0);
		fsn.listen(udpPort, tcpPort, bindInformation, replication);
		InetAddress outsideAddress = bindInformation.getListenAddresses().get(0);
		System.err.println("Listening on " + outsideAddress);
		// fsn.setMultiMap(new MultiMapDisk(dataDir, id));
	}

	public void setRamDisk()
	{
		fsn.setMultiMap(new MultiMapMemory());
	}

	public void publish(PreviewURN urn, List<String> keywords, boolean fast, AtomicBoolean publish)
			throws IOException, URISyntaxException
	{
		File fileToPublish = new File(fsn.getPublishDir(), urn.getName());
		if (!fileToPublish.exists())
			throw new RuntimeException("Hey you, copy file to the publish dir first! "
					+ "It makes no sense to publish without making the file available.");
		FuturePublish fp = fsn.addFileAndPublish(fileToPublish, urn);
		List<String> neighbors = new ArrayList<String>();
		List<BaseFuture> fdhts = new ArrayList<BaseFuture>();
		fdhts.add(fp);
		long l = System.currentTimeMillis();
		for (String keyword : keywords)
			neighbors.addAll(generateFastSS(keyword));
		System.err.println("Neighbor time: " + (System.currentTimeMillis() - l));
		l = System.currentTimeMillis();
		for (String keyword : keywords)
		{
			fdhts.add(fsn.add(Utils.makeSHAHash(keyword.toLowerCase()), domain, urn.toString()
					.getBytes()));
			Utils.sleep(fast ? 0 : 5000);
			System.err.println("keyword: " + keyword);
		}
		System.err.println("Publish 0k " + keywords.size() + " size: "
				+ (System.currentTimeMillis() - l));
		l = System.currentTimeMillis();
		for (String neighbor : neighbors)
		{
			if (!publish.get())
				return;
			fdhts.add(fsn.add(Utils.makeSHAHash(neighbor), domain, urn.toString().getBytes()));
			System.err.println("keyword: " + neighbor);
			Utils.sleep(fast ? 0 : 5000);
		}
		l = System.currentTimeMillis();
		for (BaseFuture fdht : fdhts)
		{
			if (!publish.get())
				return;
			fdht.awaitUninterruptibly();
		}
		System.err.println("DHT storing time: " + (System.currentTimeMillis() - l));
	}

	public void addFile(File f, URN urn) throws IOException, URISyntaxException
	{
		fsn.addFile(f, urn);
	}

	public static Collection<String> generateFastSS(String keyword)
	{
		StringBuilder sb = new StringBuilder(keyword.toLowerCase());
		int length = keyword.length();
		Collection<String> neighbors = new HashSet<String>(length);
		char c;
		for (int i = 0; i < length; i++)
		{
			c = sb.charAt(i);
			sb.deleteCharAt(i);
			neighbors.add(sb.toString());
			sb.insert(i, c);
		}
		return neighbors;
	}
	private boolean abort = false;

	public void abortSearch()
	{
		abort = true;
	}

	public void search(String keyword, final Map<PreviewURN, PreviewURN> result)
	{
		abort = false;
		Collection<String> neighbors = generateFastSS(keyword);
		neighbors.add(keyword.toLowerCase());
		List<FutureDHT> fdhts = new ArrayList<FutureDHT>();
		// Map<PreviewURN, PreviewURN> result = new TreeMap<PreviewURN,
		// PreviewURN>();
		fdhts.add(fsn.get(Utils.makeSHAHash(keyword), domain));
		for (String neighbor : neighbors)
		{
			System.err.println("search " + neighbor + " / " + Utils.makeSHAHash(neighbor));
			if (abort)
				break;
			fsn.get(Utils.makeSHAHash(neighbor), domain).addListener(
					new BaseFutureAdapter<FutureDHT>()
					{
						@Override
						public void operationComplete(FutureDHT fdht) throws Exception
						{
							if (fdht.isSuccess())
							{
								System.err.println("found!");
								Collection<Object> urns = fdht.getData();
								for (Object urnObj : urns)
								{
									if (urnObj instanceof byte[])
									{
										byte[] me = (byte[]) urnObj;
										try
										{
											PreviewURN urn = new PreviewURN(new String(me));
											System.err.println("found! " + urn + " / "
													+ fdht.executedOn());
											PreviewURN urn2 = result.get(urn);
											if (urn2 != null)
												urn2.incHit();
											else
												result.put(urn, urn);
											synchronized (result)
											{
												result.notify();
											}
										}
										catch (URISyntaxException e)
										{
											e.printStackTrace();
										}
									}
								}
							}
							else
							{
								System.err.println("no found!");
							}
						}
					});
		}
	}

	public void removeURN(URN urn)
	{
		fsn.removeURN(urn);
	}

	public FutureFile download(URN urn)
	{
		return fsn.downloadURN(urn);
	}

	public boolean bootstrap(InetSocketAddress bootstrapAddress)
	{
		return this.fsn.bootstrap(bootstrapAddress).awaitUninterruptibly().isSuccess();
	}

	public void shutdown()
	{
		fsn.shutdown();
	}

	public InetSocketAddress getAddress()
	{
		return fsn.getConnectionHandler().getServerNode().getSocketAddressTCP();
	}
}
