package ch.uzh.csg.filesharing.fileutils;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.util.BitSet;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import ch.uzh.csg.filesharing.URN;

public class FileChunkMapping
{
	// this shores the hash that belongs to a file. It also stores all the
	// chunks
	final private Map<BigInteger, FileChunks> fileMapping = new ConcurrentHashMap<BigInteger, FileChunks>();
	final private File incomingDir;
	final private int splitSize;

	public FileChunkMapping(File incomingDir, int splitSize)
	{
		this.incomingDir = incomingDir;
		incomingDir.mkdirs();
		this.splitSize = splitSize;
	}

	public File getCompleteFile(URN urn)
	{
		FileChunks fc = fileMapping.get(urn.getHash());
		return fc == null ? null : fc.getCompletFile();
	}

	/**
	 * Add file and hash mapping. In order to do this, we need to split the
	 * file. The split is done in a thread. So this method returns immediately
	 * and the splitting is done in background.
	 * 
	 * @param fileHash The hash that belongs to this file content
	 * @param file The file that will be split if it is large enough. Large
	 *        means larger than SplitterUtils.splitSize
	 * @throws IOException
	 */
	public void addFileMapping(File file, URN urn) throws IOException
	{
		fileMapping.put(urn.getHash(), new FileChunks(file, urn, splitSize));
	}

	public void removeMapping(BigInteger fileHash)
	{
		fileMapping.remove(fileHash);
	}

	public boolean isPartially(URN urn)
	{
		return fileMapping.containsKey(urn.getHash());
	}

	public ByteBuffer getChunk(BigInteger key, int chunkID)
	{
		FileChunks fileChunks = fileMapping.get(key);
		if (fileChunks != null)
			return fileChunks.getChunk(chunkID);
		return null;
	}

	public FileChunks getFileChunks(URN urn)
	{
		return fileMapping.get(urn.getHash());
	}

	public void addChunk(URN urn, int chunkID, ByteBuffer buffer) throws IOException
	{
		FileChunks fileChunks = fileMapping.get(urn.getHash());
		if (fileChunks == null)
		{
			fileChunks = new FileChunks(urn, incomingDir, splitSize);
			fileMapping.put(urn.getHash(), fileChunks);
		}
		fileChunks.addChunk(chunkID, buffer);
	}

	public boolean isComplete(URN urn)
	{
		FileChunks fileChunks = fileMapping.get(urn.getHash());
		if (fileChunks != null)
			return fileChunks.isComplete();
		return false;
	}

	public BitSet getChunks(BigInteger key)
	{
		FileChunks fileChunks = fileMapping.get(key);
		if (fileChunks == null)
			return null;
		return fileChunks.getChunks();
	}

	public BitSet getChunks(URN urn)
	{
		return getChunks(urn.getHash());
	}

	public double progress(URN urn)
	{
		FileChunks fileChunks = fileMapping.get(urn.getHash());
		if (fileChunks == null)
			return 0d;
		else
			return (double) fileChunks.getStoredChunkSize() / (double) fileChunks.getMaxChunks();
	}

	public Map<BigInteger, FileChunks> getFileMapping()
	{
		return fileMapping;
	}
}
