package ch.uzh.csg.filesharing.fileutils;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.BitSet;

import ch.uzh.csg.filesharing.URN;
import ch.uzh.csg.tomp2p.Utils;

public class FileChunks
{
	final private BitSet chunks;
	// TODO: need here a bitset with request that may time out...
	// final private BitSet request;
	final private File completeFile;
	final private File metaFile;
	final private int maxChunks;
	final private int splitSize;
	final private Object lock = this;
	final private URN urn;
	private int chunkCounter = 0; // stats

	/**
	 * Use this if you have the file
	 * 
	 * @param completeFile
	 * @param urn
	 * @param splitDirectory
	 * @param splitSize
	 * @throws IOException
	 */
	public FileChunks(File completeFile, URN urn, int splitSize) throws IOException
	{
		this.completeFile = completeFile;
		this.urn = urn;
		this.metaFile = new File(completeFile, completeFile.getName() + ".meta");
		this.splitSize = splitSize;
		this.maxChunks = SplitterUtils.numPieces(urn.getSize(), splitSize);
		this.chunks = new BitSet(maxChunks);
		chunks.set(0, maxChunks);
	}

	/**
	 * Use this constructor if file has to be downloaded first
	 * 
	 * @param urn
	 * @param splitDirectory
	 * @param splitSize
	 * @throws IOException
	 */
	public FileChunks(URN urn, File incomingDirectory, int splitSize) throws IOException
	{
		this.urn = urn;
		this.completeFile = new File(incomingDirectory, urn.getName());
		this.metaFile = new File(incomingDirectory, urn.getName() + ".meta");
		this.maxChunks = SplitterUtils.numPieces(urn.getSize(), splitSize);
		this.splitSize = splitSize;
		BitSet tmp = new BitSet(maxChunks);
		if (metaFile.exists())
		{
			this.chunks = readMetaData(tmp);
		}
		else
		{
			this.chunks = tmp;
			metaFile.createNewFile();
		}
	}

	public ByteBuffer getChunk(int chunkID)
	{
		FileInputStream fis = null;
		FileChannel fcin = null;
		try
		{
			long offset = chunkID * splitSize;
			synchronized (lock)
			{
				fis = new FileInputStream(completeFile);
				fcin = fis.getChannel();
				if (chunkID < maxChunks - 1)
					return fcin.map(FileChannel.MapMode.READ_ONLY, offset, splitSize);
				else
					return fcin.map(FileChannel.MapMode.READ_ONLY, offset, completeFile.length()
							- offset);
			}
		}
		catch (IOException e)
		{
			e.printStackTrace();
			return null;
		}
		finally
		{
			synchronized (lock)
			{
				Utils.bestEffortclose(fis, fcin);
			}
		}
	}

	public void addChunk(int chunkID, ByteBuffer buffer) throws IOException
	{
		synchronized (lock)
		{
			if (!isComplete())
			{
				if (!chunks.get(chunkID))
				{
					System.err.print("~");
					chunkCounter++;
					if (chunkCounter % 100 == 0)
						System.err.print("\n");
				}
				else
					System.err.print("^");
				chunks.set(chunkID);
				writeData(chunkID, buffer);
				updateMetaData();
				if (isCompleting())
				{
					System.err.println("COMPLETING " + completeFile);
					metaFile.delete();
				}
			}
		}
	}

	private void writeData(int chunkID, ByteBuffer buffer)
	{
		RandomAccessFile raf = null;
		FileChannel fout = null;
		try
		{
			raf = new RandomAccessFile(completeFile, "rw");
			fout = raf.getChannel();
			long offset = chunkID * splitSize;
			System.err.println("Write chunk " + chunkID + " offset=" + offset + " size="
					+ buffer.limit());
			fout.position(offset).write(buffer);
			fout.force(true);
		}
		catch (IOException e)
		{
			e.printStackTrace();
		}
		finally
		{
			Utils.bestEffortclose(raf, fout);
		}
	}

	public boolean isComplete()
	{
		return isCompleting() && !metaFile.exists();
	}

	public boolean isCompleting()
	{
		return chunks.cardinality() == maxChunks;
	}

	public File getCompletFile()
	{
		return completeFile;
	}

	public BitSet getChunks()
	{
		return chunks;
	}

	/*
	 * public List<Integer> getMissingChunks() { synchronized (lock) {
	 * List<Integer> retVal = new ArrayList<Integer>(); for (int i = 0; i <
	 * maxChunks; i++) { if (!chunks.get(i)) retVal.add(i); } return retVal; } }
	 * 
	 * public List<Integer> getStoredChunks() { synchronized (lock) {
	 * List<Integer> retVal = new ArrayList<Integer>(); for (int i = 0; i <
	 * maxChunks; i++) { if (chunks.get(i)) retVal.add(i); } return retVal; } }
	 */
	public int getStoredChunkSize()
	{
		return chunks.cardinality();
	}

	public int getMissingChunkSize()
	{
		return maxChunks - chunks.cardinality();
	}

	public int getMaxChunks()
	{
		return maxChunks;
	}

	public RandomBitSet diff(BitSet otherChunks)
	{
		synchronized (lock)
		{
			RandomBitSet tmp = new RandomBitSet(otherChunks);
			tmp.xor(chunks);
			tmp.and(otherChunks);
			return tmp;
		}
	}

	public void resetChunkCounter()
	{
		chunkCounter = 0;
	}

	public int getChunkCounter()
	{
		return chunkCounter;
	}

	private BitSet readMetaData(BitSet old)
	{
		FileInputStream in = null;
		ObjectInputStream s = null;
		try
		{
			in = new FileInputStream(metaFile);
			s = new ObjectInputStream(in);
			URN urn = (URN) s.readObject();
			return (BitSet) s.readObject();
		}
		catch (IOException e)
		{
			return old;
		}
		catch (ClassNotFoundException e)
		{
			return old;
		}
		finally
		{
			Utils.bestEffortclose(in, s);
		}
	}

	private void updateMetaData() throws IOException
	{
		FileOutputStream out = new FileOutputStream(metaFile);
		ObjectOutputStream s = new ObjectOutputStream(out);
		try
		{
			s.writeObject(urn);
			s.writeObject(chunks);
			s.flush();
		}
		finally
		{
			Utils.bestEffortclose(out, s);
		}
	}
}
