package ch.uzh.csg.filesharing.fileutils;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.List;
import java.util.Observable;

import ch.uzh.csg.tomp2p.Utils;

public class SplitterThread extends Observable implements Runnable
{
	final private long fileSize;
	final private long splitSize;
	final private File sourceFile;
	final private File destinationFolder;
	final private long noOfPieces;
	final private boolean digest;
	final private List<BigInteger> digestArray = new ArrayList<BigInteger>();
	final private List<File> splits = new ArrayList<File>();
	//
	private int splitCounter = 0;
	private boolean finishedSplitting = false;

	public SplitterThread(File sourceFile, File destinationFolder, long splitSize)
	{
		this(sourceFile, destinationFolder, splitSize, false);
	}

	public SplitterThread(File sourceFile, File destinationFolder, long splitSize, boolean digest)
	{
		this.destinationFolder = destinationFolder;
		this.sourceFile = sourceFile;
		this.fileSize = sourceFile.length();
		this.splitSize = splitSize;
		this.digest = digest;
		this.noOfPieces = SplitterUtils.numPieces(fileSize, splitSize);
		if (noOfPieces == 0)
			throw new IllegalArgumentException("Nr of splits can never be 0. Filesize=" + fileSize
					+ ", splitSize=" + splitSize);
	}

	public long getNoOfPieces()
	{
		return noOfPieces;
	}

	public int getSplitCounter()
	{
		return splitCounter;
	}

	public boolean hasFinishedSplitting()
	{
		return finishedSplitting;
	}

	private void incrementProgressCount(File piece)
	{
		setChanged();
		notifyObservers(piece);
		splitCounter++;
	}

	public void split()
	{
		SplitterUtils.getThreadPool().execute(this);
	}

	public void run()
	{
		try
		{
			splitFile();
		}
		catch (Exception e)
		{
			System.err.println("Error while splitting " + e);
			e.printStackTrace();
		}
	}

	public void splitFile() throws IOException
	{
		FileInputStream fis = null;
		FileOutputStream fos = null;
		FileChannel fcin = null;
		FileChannel fcout = null;
		try
		{
			fis = new FileInputStream(sourceFile);
			fcin = fis.getChannel();
			// long fileSize = fcin.size();
			long offset = 0L;
			long chunkSize = sourceFile.length() < splitSize ? sourceFile.length() : splitSize;
			for (int i = 0; i < noOfPieces; i++)
			{
				File filepiece = new File(destinationFolder, sourceFile.getName() + "_" + i
						+ ".split");
				if (!filepiece.exists())
				{
					filepiece.getParentFile().mkdirs();
					filepiece.createNewFile();
					fos = new FileOutputStream(filepiece);
					fcout = fos.getChannel();
					ByteBuffer mbb = fcin.map(FileChannel.MapMode.READ_ONLY, offset, chunkSize);
					fcout.write(mbb);
					if (digest)
					{
						mbb.flip();
						digestArray.add(SplitterUtils.calculateDigest(mbb));
					}
					fcout.close();
					fos.close();
				}
				offset += splitSize;
				if (fileSize - offset < chunkSize)
				{ // last chunk.
					chunkSize = fileSize - offset;
				}
				splits.add(filepiece);
				incrementProgressCount(filepiece);
			}
		}
		finally
		{
			Utils.bestEffortclose(fcout, fos, fcin, fis);
			finishedSplitting = true;
			incrementProgressCount(null);
		}
	}

	public List<File> getSplitFiles()
	{
		return splits;
	}

	public File getCurrentFile()
	{
		if (splits.size() > 0)
			return splits.get(splits.size() - 1);
		return null;
	}
}
