package ch.uzh.csg.filesharing;
import java.io.File;
import java.io.IOException;
import java.math.BigInteger;
import java.net.URISyntaxException;
import java.util.concurrent.atomic.AtomicBoolean;

import org.junit.Assert;
import org.junit.Test;

import ch.uzh.csg.filesharing.FileSharingNode.Strategy;
import ch.uzh.csg.filesharing.fileutils.FileUtils;
import ch.uzh.csg.filesharing.future.FutureFile;
import ch.uzh.csg.filesharing.future.FuturePublish;
import ch.uzh.csg.tomp2p.Replication;
import ch.uzh.csg.tomp2p.Utils;
import ch.uzh.csg.tomp2p.common.BaseFutureAdapter;
import ch.uzh.csg.tomp2p.common.BaseFutureListener;
import ch.uzh.csg.tomp2p.connection.BindInformation;
import ch.uzh.csg.tomp2p.connection.BindInformation.Protocol;

public class TestPSH
{
	private FileSharingNode[] prep(int p2pId, Strategy st) throws IOException
	{
		BindInformation bindInformation = new BindInformation();
		bindInformation.addListenProtocolHint(Protocol.IPv4);
		FileSharingNode[] fileSharingNodes = new FileSharingNode[20];
		for (int i = 0; i < fileSharingNodes.length; i++)
		{
			BigInteger id = Utils.makeSHAHash("" + i);
			fileSharingNodes[i] = new FileSharingNode(p2pId, id, st, 1024 * 3, 1024, null, null,
					new File("/tmp/a"), new File("/tmp/b"), 0, 0);
			fileSharingNodes[i].listen(1234, 1234, bindInformation, new Replication());
			if (i != 0)
				fileSharingNodes[i].bootstrap(fileSharingNodes[0]).awaitUninterruptibly();
		}
		return fileSharingNodes;
	}

	private void close(FileSharingNode[] fileSharingNodes)
	{
		for (FileSharingNode node : fileSharingNodes)
			node.shutdown();
	}

	@Test
	public void testTFT1() throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(1, Strategy.TFT);
		File f = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.TEN);
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f);
		fp.awaitUninterruptibly();
		final AtomicBoolean tmp = new AtomicBoolean(false);
		final FutureFile fs = fileSharingNodes[1].downloadURN(new URN(f), 3000);
		fs.awaitUninterruptibly();
		fs.addListener(new BaseFutureListener<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("success: " + future.isSuccess());
				System.err.println(future.getFailedReason());
				System.err.println("progress: " + prog);
				tmp.set(0.3d == prog);
				System.err.println("done1");
			}

			@Override
			public void exceptionCaught(Throwable t) throws Exception
			{
				t.printStackTrace();
			}
		});
		Assert.assertEquals(true, tmp.get());
		System.err.println("done2");
		close(fileSharingNodes);
	}

	@Test
	public void testTFT2() throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(2, Strategy.TFT);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.TEN);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ONE);
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		//
		FutureFile fs1 = fileSharingNodes[1].downloadURN(new URN(f1));
		fs1.addListener(new BaseFutureListener<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
			}

			@Override
			public void exceptionCaught(Throwable t) throws Exception
			{
				t.printStackTrace();
			}
		});
		FutureFile fs2 = fileSharingNodes[0].downloadURN(new URN(f2));
		fs2.addListener(new BaseFutureListener<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
			}

			@Override
			public void exceptionCaught(Throwable t) throws Exception
			{
				t.printStackTrace();
			}
		});
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		Assert.assertEquals(1d, fs1.getProgress(), 0.001d);
		Assert.assertEquals(1d, fs2.getProgress(), 0.001d);
		Assert.assertEquals(Utils.makeSHAHash(f1), Utils.makeSHAHash(fs1.getFile()));
		Assert.assertEquals(Utils.makeSHAHash(f2), Utils.makeSHAHash(fs2.getFile()));
		close(fileSharingNodes);
	}

	@Test
	public void testTFT3() throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(3, Strategy.TFT);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.TEN);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ONE);
		File f3 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 99, BigInteger.ZERO);
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[2].addFileAndPublish(f3);
		fp.awaitUninterruptibly();
		//
		final FutureFile fs1 = fileSharingNodes[1].downloadURN(new URN(f3), 5000);
		fs1.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
			}
		});
		final FutureFile fs2 = fileSharingNodes[0].downloadURN(new URN(f2), 5000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
			}
		});
		final FutureFile fs3 = fileSharingNodes[2].downloadURN(new URN(f1), 5000);
		fs3.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress3: " + prog);
			}
		});
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		fs3.awaitUninterruptibly();
		Assert.assertEquals(0.3d, fs1.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs2.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs3.getProgress(), 0.001d);
		close(fileSharingNodes);
	}

	@Test
	public void testTFT4() throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(4, Strategy.TFT);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.TEN);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ONE);
		File f3 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 99, BigInteger.ZERO);
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[2].addFileAndPublish(f3);
		fp.awaitUninterruptibly();
		final FutureFile fs1 = fileSharingNodes[2].downloadURN(new URN(f2), 5000);
		fs1.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
				// Utils.sleep(Long.MAX_VALUE);
			}
		});
		final FutureFile fs2 = fileSharingNodes[0].downloadURN(new URN(f2), 5000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
				// Utils.sleep(Long.MAX_VALUE);
			}
		});
		final FutureFile fs3 = fileSharingNodes[2].downloadURN(new URN(f1), 5000);
		fs3.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress3: " + prog);
			}
		});
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		fs3.awaitUninterruptibly();
		Assert.assertTrue(fs1.getProgress() <= 0.6d && fs1.getProgress() >= 0.3d);
		Assert.assertTrue(fs2.getProgress() <= 0.6d && fs1.getProgress() >= 0.3d);
		Assert.assertTrue(fs3.getProgress() <= 0.6d && fs1.getProgress() >= 0.3d);
		close(fileSharingNodes);
	}

	@Test
	public void testPSH2_1() throws IOException, URISyntaxException
	{
		testPSH1(5, Strategy.PSH2);
	}

	@Test
	public void testPSHR_1() throws IOException, URISyntaxException
	{
		testPSH1(6, Strategy.PSHR);
	}

	private void testPSH1(int p2pId, Strategy strategy) throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(p2pId, strategy);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.TEN);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ONE);
		File f3 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 99, BigInteger.ZERO);
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[2].addFileAndPublish(f3);
		fp.awaitUninterruptibly();
		//
		final FutureFile fs1 = fileSharingNodes[1].downloadURN(new URN(f3), 10000);
		fs1.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
				// Utils.sleep(Long.MAX_VALUE);
			}
		});
		final FutureFile fs2 = fileSharingNodes[0].downloadURN(new URN(f2), 10000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
				// Utils.sleep(Long.MAX_VALUE);
			}
		});
		final FutureFile fs3 = fileSharingNodes[2].downloadURN(new URN(f1), 10000);
		fs3.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress3: " + prog);
			}
		});
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		fs3.awaitUninterruptibly();
		Assert.assertEquals(1d, fs1.getProgress(), 0.001d);
		Assert.assertEquals(1d, fs2.getProgress(), 0.001d);
		Assert.assertEquals(1d, fs3.getProgress(), 0.001d);
		Assert.assertEquals(Utils.makeSHAHash(f3), Utils.makeSHAHash(fs1.getFile()));
		Assert.assertEquals(Utils.makeSHAHash(f2), Utils.makeSHAHash(fs2.getFile()));
		Assert.assertEquals(Utils.makeSHAHash(f1), Utils.makeSHAHash(fs3.getFile()));
		close(fileSharingNodes);
	}

	@Test
	public void testPSH2_2() throws IOException, URISyntaxException
	{
		testPSH2(7, Strategy.PSH2);
	}

	@Test
	public void testPSHR_2() throws IOException, URISyntaxException
	{
		testPSH2(8, Strategy.PSHR);
	}

	private void testPSH2(int p2pId, Strategy strategy) throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(p2pId, strategy);
		File f0 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 66, BigInteger.TEN);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.ONE);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ZERO);
		File f3 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 99, new BigInteger("2"));
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f0);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[2].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[3].addFileAndPublish(f3);
		fp.awaitUninterruptibly();
		//
		final FutureFile fs0 = fileSharingNodes[0].downloadURN(new URN(f1), 10000);
		fs0.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
			}
		});
		final FutureFile fs1 = fileSharingNodes[1].downloadURN(new URN(f2), 10000);
		fs1.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
			}
		});
		final FutureFile fs2 = fileSharingNodes[2].downloadURN(new URN(f3), 10000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress3: " + prog);
			}
		});
		final FutureFile fs3 = fileSharingNodes[3].downloadURN(new URN(f0), 10000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress4: " + prog);
			}
		});
		fs0.awaitUninterruptibly();
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		fs3.awaitUninterruptibly();
		Assert.assertEquals(0.3d, fs0.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs1.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs2.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs3.getProgress(), 0.001d);
		close(fileSharingNodes);
	}

	@Test
	public void testPSH2_3() throws IOException, URISyntaxException
	{
		testPSH3(9, Strategy.PSH2);
	}

	@Test
	public void testPSHR_3() throws IOException, URISyntaxException
	{
		testPSH3(10, Strategy.PSHR);
	}

	private void testPSH3(int p2pId, Strategy strategy) throws IOException, URISyntaxException
	{
		FileSharingNode[] fileSharingNodes = prep(p2pId, strategy);
		File f0 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 66, BigInteger.TEN);
		File f1 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 77, BigInteger.ONE);
		File f2 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 88, BigInteger.ZERO);
		File f3 = FileUtils.generateRadomFile(10 * 1024, new File("/tmp"), 99, new BigInteger("2"));
		//
		FuturePublish fp = fileSharingNodes[0].addFileAndPublish(f0);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[1].addFileAndPublish(f1);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[2].addFileAndPublish(f2);
		fp.awaitUninterruptibly();
		fp = fileSharingNodes[3].addFileAndPublish(f3);
		fp.awaitUninterruptibly();
		//
		final FutureFile fs0 = fileSharingNodes[0].downloadURN(new URN(f1), 5000);
		fs0.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress1: " + prog);
			}
		});
		final FutureFile fs1 = fileSharingNodes[1].downloadURN(new URN(f2), 5000);
		fs1.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress2: " + prog);
			}
		});
		final FutureFile fs2 = fileSharingNodes[2].downloadURN(new URN(f3), 5000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress3: " + prog);
			}
		});
		final FutureFile fs3 = fileSharingNodes[3].downloadURN(new URN(f1), 5000);
		fs2.addListener(new BaseFutureAdapter<FutureFile>()
		{
			@Override
			public void operationComplete(FutureFile future)
			{
				double prog = future.getProgress();
				System.err.println("progress4: " + prog);
			}
		});
		fs0.awaitUninterruptibly();
		fs1.awaitUninterruptibly();
		fs2.awaitUninterruptibly();
		fs3.awaitUninterruptibly();
		// download from two locations
		Assert.assertEquals(1d, fs1.getProgress(), 0.001d);
		Assert.assertEquals(1d, fs2.getProgress(), 0.001d);
		Assert.assertEquals(1d, fs3.getProgress(), 0.001d);
		// this is 0.6, but due to slow updates, we can notified too late, so
		// its 0.3
		// Assert.assertEquals(0.6d, fs0.getProgress(), 0.001d);
		Assert.assertEquals(0.3d, fs0.getProgress(), 0.001d);
		close(fileSharingNodes);
	}
}
